# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# +
from functools import reduce

import torch
import torch.nn as nn
from torch.hub import load_state_dict_from_url
from .base_model import BaseModel

# -

from collections import defaultdict


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding."""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=3,
                     stride=stride,
                     padding=1,
                     bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.active = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.active(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.active(out)

        return out

    def get_flops(self, inp):
        flops = 0
        output = self.conv1(inp)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv1.in_channels * self.conv1.out_channels * 9 *
                  active_elements_count)
        flops += 4 * output.numel()  # relu + bn
        output = self.conv2(output)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv2.in_channels * self.conv2.out_channels * 9 *
                  active_elements_count)
        flops += 5 * output.numel()  # relu + bn + residual

        if self.downsample is not None:
            r_output = self.downsample[0](inp)
            active_elements_count = r_output.shape[0] * reduce(
                lambda x, y: x * y, r_output.shape[2:])
            flops += (self.downsample[0].in_channels *
                      self.downsample[0].out_channels * active_elements_count)
            flops += 2 * r_output.numel()
        return flops


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes,
                               planes,
                               kernel_size=3,
                               stride=stride,
                               padding=1,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes,
                               planes * self.expansion,
                               kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.active = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.active(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.active(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.active(out)

        return out


class ResNetCifar(BaseModel):

    def __init__(self, block, layers, width=1, num_classes=1000, insize=32):
        super(ResNetCifar, self).__init__()
        self.inplanes = 16
        self.insize = insize
        self.layers_size = layers
        self.num_classes = num_classes
        self.width = width
        self.conv1 = nn.Conv2d(3,
                               16,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.prev_module = defaultdict()
        self.prev_module[self.bn1] = None
        self.active = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, 16 * width, layers[0])
        self.layer2 = self._make_layer(block, 32 * width, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 64 * width, layers[2], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)
        self.fc = nn.Linear(64 * width, num_classes)
        self.init_weights()

        assert block is BasicBlock
        prev = self.bn1
        for l_block in [self.layer1, self.layer2, self.layer3]:
            for b in l_block:
                self.prev_module[b.bn1] = prev
                self.prev_module[b.bn2] = b.bn1
                if b.downsample is not None:
                    self.prev_module[b.downsample[1]] = prev
                prev = b.bn2

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            conv_module = nn.Conv2d(
                self.inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=stride,
                bias=False,
            )
            bn_module = nn.BatchNorm2d(planes * block.expansion)
            if hasattr(bn_module, 'is_imp'):
                bn_module.is_imp = True
            downsample = nn.Sequential(conv_module, bn_module)

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.active(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x

    def get_bn_layers(self):
        bn_layers = []
        for l_blocks in [self.layer1, self.layer2, self.layer3]:
            for b in l_blocks:
                m1, m2 = b.bn1, b.bn2
                bn_layers.append([m1, m2])
        return bn_layers


class ResNet(BaseModel):

    def __init__(
        self,
        block,
        layers,
        width=1,
        num_classes=1000,
        produce_vectors=False,
        init_weights=True,
        insize=32,
    ):
        super(ResNet, self).__init__()
        self.layers_size = layers
        self.num_classes = num_classes
        self.insize = insize
        self.produce_vectors = produce_vectors
        self.block_type = block.__class__.__name__
        self.inplanes = 64
        if insize < 128:
            self.conv1 = nn.Conv2d(3,
                                   64,
                                   kernel_size=3,
                                   stride=1,
                                   padding=1,
                                   bias=False)
        else:
            self.conv1 = nn.Conv2d(3,
                                   64,
                                   kernel_size=7,
                                   stride=2,
                                   padding=3,
                                   bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.prev_module = defaultdict()
        self.prev_module[self.bn1] = None
        self.active = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64 * width, layers[0])
        self.layer2 = self._make_layer(block, 128 * width, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256 * width, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512 * width, layers[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)  # Global Avg Pool
        self.fc = nn.Linear(512 * block.expansion * width, num_classes)

        self.init_weights()

        for l in [self.layer1, self.layer2, self.layer3, self.layer4]:
            for b in l.children():
                downs = (next(b.downsample.children())
                         if b.downsample is not None else None)

        # assert block is Bottleneck
        if block is Bottleneck:
            prev = self.bn1
            for l_block in [
                    self.layer1, self.layer2, self.layer3, self.layer4
            ]:
                for b in l_block:
                    self.prev_module[b.bn1] = prev
                    self.prev_module[b.bn2] = b.bn1
                    self.prev_module[b.bn3] = b.bn2
                    if b.downsample is not None:
                        self.prev_module[b.downsample[1]] = prev
                    prev = b.bn3

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            conv_module = nn.Conv2d(
                self.inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=stride,
                bias=False,
            )
            bn_module = nn.BatchNorm2d(planes * block.expansion)
            if hasattr(bn_module, 'is_imp'):
                bn_module.is_imp = True
            downsample = nn.Sequential(conv_module, bn_module)

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.active(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        feature_vectors = x.view(x.size(0), -1)
        x = self.fc(feature_vectors)

        if self.produce_vectors:
            return x, feature_vectors
        else:
            return x

    def get_bn_layers(self):
        bn_layers = []
        for l_blocks in [self.layer1, self.layer2, self.layer3, self.layer4]:
            for b in l_blocks:
                if self.block_type == 'Bottleneck':
                    m1, m2, m3 = b.bn1, b.bn2, b.bn3
                    bn_layers.append([m1, m2, m3])
                else:
                    m1, m2 = b.bn1, b.bn2
                    bn_layers.append([m1, m2])
        return bn_layers

    def get_flops(self, insize):
        flops = 0
        inp = torch.rand(1, 3, insize, insize).to(self.conv1.weight.device)
        output = self.conv1(inp)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv1.in_channels * self.conv1.out_channels * 9 *
                  active_elements_count)
        output = self.bn1(output)
        flops += 2 * output.numel()
        output = self.active(output)
        flops += 2 * output.numel()
        output = self.maxpool(output)
        flops += 8 * output.numel()

        for l_block in [self.layer1, self.layer2, self.layer3, self.layer4]:
            for b in l_block:
                flops += b.get_flops(output)
                output = b(output)

        ops = output.shape[-1] * output.shape[-2]
        output = self.avgpool(output)
        flops += ops * output.numel()

        flops += (1 + self.fc.in_features) * self.fc.out_features
        return flops


def make_wide_resnet(num_classes, insize):
    model = ResNetCifar(BasicBlock, [4, 4, 4],
                        width=12,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet20(num_classes, insize):
    model = ResNetCifar(BasicBlock, [3, 3, 3],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet32(num_classes, insize):
    model = ResNetCifar(BasicBlock, [5, 5, 5],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet50(num_classes, insize):
    model = ResNet(Bottleneck, [3, 4, 6, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet26(num_classes, insize):
    model = ResNet(BasicBlock, [3, 3, 3, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet34(num_classes, insize):
    model = ResNet(BasicBlock, [3, 4, 6, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet18(num_classes, insize):
    model = ResNet(BasicBlock, [2, 2, 2, 2],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet10(num_classes, insize):
    model = ResNet(BasicBlock, [1, 1, 1, 1],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet56(num_classes, insize):
    model = ResNetCifar(BasicBlock, [9, 9, 9],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet101(num_classes, insize):
    model = ResNet(Bottleneck, [3, 4, 23, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet110(num_classes, insize):
    model = ResNetCifar(BasicBlock, [18, 18, 18],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet152(num_classes, insize):
    model = ResNet(Bottleneck, [3, 8, 36, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def get_resnet_model(model, num_classes, insize, pretrained):
    """Returns the requested model, ready for training/pruning with the
    specified method.

    :param model: str, either wrn or r50
    :param num_classes: int, num classes in the dataset
    :return: A prunable ResNet model
    """
    if model == 'wrn':
        net = make_wide_resnet(num_classes, insize)
        pretrained_weights = None

    elif model == 'resnet18':
        net = make_resnet18(num_classes, insize)
        pretrained_weights = 'https://download.pytorch.org/models/resnet18-f37072fd.pth'
    elif model == 'resnet20':
        net = make_resnet20(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet32':
        net = make_resnet32(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet50':
        net = make_resnet50(num_classes, insize)
        pretrained_weights = 'https://download.pytorch.org/models/resnet50-11ad3fa6.pth'
    elif model == 'resnet56':
        net = make_resnet56(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet101':
        net = make_resnet101(num_classes, insize)
        pretrained_weights = (
            'https://download.pytorch.org/models/resnet101-cd907fc2.pth')
    elif model == 'resnet110':
        net = make_resnet110(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet152':
        net = make_resnet152(num_classes, insize)
        pretrained_weights = (
            'https://download.pytorch.org/models/resnet152-f82ba261.pth')
    elif model == 'resnet34':
        net = make_resnet34(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet26':
        net = make_resnet26(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet10':
        net = make_resnet10(num_classes, insize)
        pretrained_weights = None
    if pretrained:
        if pretrained_weights != None:
            weights = load_state_dict_from_url(pretrained_weights,
                                               progress=True)
            net.load_state_dict(weights, strict=False)
        else:
            print('pretrained weights not available')
    return net



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import torch.nn.init as init
from tqdm import tqdm_notebook
from ..algorithms import BaseAlgorithm

__all__ = [
    'BaseQuantization',
    'StraightThrough',
    'RoundSTE',
    'Conv2dFunctor',
    'LinearFunctor',
    'FoldBN',
]


class BaseQuantization(BaseAlgorithm):
    """Base class for quantization algorithms."""

    def __init__(self, **kwargs):
        super(BaseQuantization, self).__init__(**kwargs)
        pass

    def quantize(self, model, dataloaders, method, **kwargs):
        pass

    def round_ste(x: torch.Tensor):
        """Implement Straight-Through Estimator for rounding operation."""
        return (x.round() - x).detach() + x

    def get_calib_samples(self, train_loader, num_samples):
        """Get calibration-set samples for finetuning weights and clipping
        parameters."""
        calib_data = []
        for batch in train_loader:
            calib_data.append(batch[0])
            if len(calib_data) * batch[0].size(0) >= num_samples:
                break
        return torch.cat(calib_data, dim=0)[:num_samples]

    def absorb_bn(self, module, bn_module):
        w = module.weight.data
        if module.bias is None:
            zeros = torch.Tensor(module.out_channels).zero_().type(w.type())
            module.bias = nn.Parameter(zeros)
        b = module.bias.data
        invstd = bn_module.running_var.clone().add_(bn_module.eps).pow_(-0.5)
        w.mul_(invstd.view(w.size(0), 1, 1, 1).expand_as(w))
        b.add_(-bn_module.running_mean).mul_(invstd)

        if bn_module.affine:
            w.mul_(bn_module.weight.data.view(w.size(0), 1, 1, 1).expand_as(w))
            b.mul_(bn_module.weight.data).add_(bn_module.bias.data)

        bn_module.register_buffer('running_mean',
                                  torch.zeros(module.out_channels).cuda())
        bn_module.register_buffer('running_var',
                                  torch.ones(module.out_channels).cuda())
        bn_module.register_parameter('weight', None)
        bn_module.register_parameter('bias', None)
        bn_module.affine = False

    def is_bn(self, m):
        return isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d)

    def is_absorbing(self, m):
        return (isinstance(m, nn.Conv2d) and m.groups == 1) or isinstance(
            m, nn.Linear)

    def search_absorbe_bn(self, model):
        prev = None
        for m in model.children():
            if self.is_bn(m) and self.is_absorbing(prev):
                m.absorbed = True
                self.absorb_bn(prev, m)
            self.search_absorbe_bn(m)
            prev = m


class StraightThrough(nn.Module):
    """Used to place an identity function in place of a non-differentail
    operator for gradient calculation."""

    def __int__(self):
        super().__init__()
        pass

    def forward(self, input):
        return input


class RoundSTE(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input):
        output = torch.round(input)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output


class Conv2dFunctor:

    def __init__(self, conv2d):
        self.conv2d = conv2d

    def __call__(self, *input, weight, bias):
        res = torch.nn.functional.conv2d(*input, weight, bias,
                                         self.conv2d.stride,
                                         self.conv2d.padding,
                                         self.conv2d.dilation,
                                         self.conv2d.groups)
        return res


class LinearFunctor:

    def __init__(self, linear):
        self.linear = linear

    def __call__(self, *input, weight, bias):
        res = torch.nn.functional.linear(*input, weight, bias)
        return res


# TODO : To migrate all BN-layer folding function calls to the ones defined inside BaseQuantization class
class FoldBN:
    """Used to fold batch norm to prev linear or conv layer which helps reduce
    comutational overhead during quantization."""

    def __init__(self):
        pass

    def _fold_bn(self, conv_module, bn_module):
        w = conv_module.weight.data
        y_mean = bn_module.running_mean
        y_var = bn_module.running_var
        safe_std = torch.sqrt(y_var + bn_module.eps)
        w_view = (conv_module.out_channels, 1, 1, 1)
        if bn_module.affine:
            weight = w * (bn_module.weight / safe_std).view(w_view)
            beta = bn_module.bias - bn_module.weight * y_mean / safe_std
            if conv_module.bias is not None:
                bias = bn_module.weight * conv_module.bias / safe_std + beta
            else:
                bias = beta
        else:
            weight = w / safe_std.view(w_view)
            beta = -y_mean / safe_std
            if conv_module.bias is not None:
                bias = conv_module.bias / safe_std + beta
            else:
                bias = beta
        return weight, bias

    def fold_bn_into_conv(self, conv_module, bn_module):
        w, b = self._fold_bn(conv_module, bn_module)
        if conv_module.bias is None:
            conv_module.bias = nn.Parameter(b)
        else:
            conv_module.bias.data = b
        conv_module.weight.data = w
        # set bn running stats
        bn_module.running_mean = bn_module.bias.data
        bn_module.running_var = bn_module.weight.data**2

    def is_bn(self, m):
        return isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d)

    def is_absorbing(self, m):
        return (isinstance(m, nn.Conv2d)) or isinstance(m, nn.Linear)

    def search_fold_and_remove_bn(self, model: nn.Module):
        """Method to recursively search for batch norm layers, absorb them into
        the previous linear or conv layers, and set it to an identity layer."""
        model.eval()
        prev = None
        for n, m in model.named_children():
            if self.is_bn(m) and self.is_absorbing(prev):
                self.fold_bn_into_conv(prev, m)
                # set the bn module to straight through
                setattr(model, n, StraightThrough())
            elif self.is_absorbing(m):
                prev = m
            else:
                prev = self.search_fold_and_remove_bn(m)
        return prev



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import time
import torch
from codecarbon import OfflineEmissionsTracker


class BaseBenchmark:
    """Class to benchmark a model on a given device.

    Args:
        model (torch.nn.Module): The model to benchmark.
        batch_size (int): The batch size to use for benchmarking.
        input_size (int): The input size to use for benchmarking.
        device_name (str): The device to use for benchmarking.
        warmup_iters (int): The number of warmup iterations to use for benchmarking.
        num_iters (int): The number of iterations to use for benchmarking.
    """

    def __init__(self,
                 model,
                 batch_size,
                 input_size,
                 device_name,
                 warmup_iters=50,
                 num_iters=50):
        self.model = model
        self.batch_size = batch_size
        self.input_size = input_size
        self.device = torch.device(device_name)
        self.warmup_iters = warmup_iters
        self.num_iters = num_iters
        self.model.to(self.device)

    def measure_parameters(self):
        """Returns the number of trainable and total parameters in the
        model."""
        trainable_params = sum(p.numel() for p in self.model.parameters()
                               if p.requires_grad)
        total_params = sum(p.numel() for p in self.model.parameters())

        return trainable_params, total_params

    def measure_size(self):
        """Returns the size of the model in bytes."""
        param_size = 0
        for param in self.model.parameters():
            param_size += param.nelement() * param.element_size()
        buffer_size = 0
        for buffer in self.model.buffers():
            buffer_size += buffer.nelement() * buffer.element_size()

        return param_size + buffer_size

    def measure_throughput(self,
                           batch_size,
                           input_size,
                           warmup_iters=50,
                           num_iters=50):
        """Returns the throughput of the model in images per second."""
        torch.cuda.empty_cache()
        self.model.to(self.device)
        self.model.eval()

        timing = []
        inputs = torch.randn(batch_size,
                             3,
                             input_size,
                             input_size,
                             device=self.device)

        for _ in range(warmup_iters):
            self.model(inputs)

        torch.cuda.synchronize()
        for _ in range(num_iters):
            start = time.time()
            self.model(inputs)
            torch.cuda.synchronize()
            timing.append(time.time() - start)

        timing = torch.as_tensor(timing, dtype=torch.float32)

        return (self.batch_size / timing.mean()).item()

    def measure_flops(self, input_size):
        """Returns the number of FLOPs in the model."""
        return self.model.get_flops(input_size)

    def measure_memory(self, batch_size, input_size, num_iters):
        """Returns the peak memory utilization of the model."""
        self.model.to('cpu')
        torch.cuda.reset_peak_memory_stats(device=self.device)
        pre_mem = torch.cuda.memory_allocated(device=self.device)
        self.model.to(self.device)
        inputs = torch.randn(batch_size,
                             3,
                             input_size,
                             input_size,
                             device=self.device)
        for i in range(num_iters):
            self.model(inputs)
        max_mem = torch.cuda.max_memory_allocated(device=self.device)

        return pre_mem, max_mem

    def measure_energy(self, batch_size, input_size, num_iters=50):
        """Returns the energy consumption of the model."""
        self.model.eval()
        inputs = torch.randn(batch_size,
                             3,
                             input_size,
                             input_size,
                             device=self.device)

        with OfflineEmissionsTracker(country_iso_code='IND',
                                     log_level='error') as tracker:
            for i in range(num_iters):
                self.model(inputs)

        return (
            tracker._total_energy.kWh,
            tracker._total_cpu_energy.kWh,
            tracker._total_gpu_energy.kWh,
            tracker._total_ram_energy.kWh,
        )

    def benchmark(self, verbose=True):
        """Prints the benchmark results and returns them as a tuple."""
        _, total_params = self.measure_parameters()
        model_size = self.measure_size()
        throughput = self.measure_throughput(self.batch_size, self.input_size,
                                             self.warmup_iters, self.num_iters)

        flops = self.measure_flops(self.input_size)

        pre_mem, max_mem = self.measure_memory(self.batch_size,
                                               self.input_size, self.num_iters)
        peak_utilization = max_mem - pre_mem

        energy_consumption, _, _, _ = self.measure_energy(
            self.batch_size, self.input_size, self.num_iters)

        if verbose:
            model_name = self.model.__class__.__name__
            print(f'{model_name}\n' + '-' * (len(model_name) + 2))
            print(f'Number of parameters: {total_params}')
            print(f'Model size: {(model_size/1024**2):.2f} MB')
            print(f'Model FLOPs: {flops}')
            print(f'Throughput: {throughput:.2f} img/s')
            print(
                f'Peak GPU Memory Utilization: {(peak_utilization/1024**2):.2f} MB'
            )
            print(f'Energy Consumption: {energy_consumption:.9f} kWh')

        return (
            total_params,
            model_size,
            throughput,
            flops,
            peak_utilization,
            energy_consumption,
        )


class ModelBenchmark(BaseBenchmark):

    def __init__(self,
                 model,
                 batch_size,
                 input_size,
                 device_name,
                 warmup_iters=50,
                 num_iters=50):
        super().__init__(model, batch_size, input_size, device_name,
                         warmup_iters, num_iters)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import sys
import numpy as np
import time, datetime
import torch
import logging
import torch.nn as nn
import torch.utils
import torch.backends.cudnn as cudnn
import torch.utils.data.distributed

import timm
from .utils import DistributionLoss
from trailmet.utils import (
    AverageMeter,
    save_checkpoint,
    accuracy,
    CrossEntropyLabelSmooth,
)
import torchvision.models as models
from timm.utils.agc import adaptive_clip_grad
import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd

logger = logging.getLogger(__name__)


class BNNBN:

    def __init__(self, model, dataloaders, **kwargs):
        self.train_loader = dataloaders['train']
        self.val_loader = dataloaders['val']
        self.test_loader = dataloaders['test']
        self.model_student = model
        self.kwargs = kwargs
        self.dataset = self.kwargs.get('DATASET', 'c100')
        self.num_classes = self.kwargs.get('num_classes', 100)
        self.loss_type = self.kwargs.get('loss_type', 'kd')
        self.teacher = self.kwargs.get('teacher', 'resnet34')
        self.teacher_weight = self.kwargs.get('teacher_weight', '')
        self.label_smooth = self.kwargs.get('label_smooth', '0.1')
        self.pretrained = self.kwargs.get('pretrained', '')
        self.resume = self.kwargs.get('resume', 'False')
        self.save = self.kwargs.get('save', '')
        self.epochs = self.kwargs.get('epoch', '120')
        self.agc = self.kwargs.get('agc', 'True')
        self.clip_value = self.kwargs.get('clip_value', '0.04')
        self.weight_decay = self.kwargs.get('weight_decay', '0')
        self.learning_rate = self.kwargs.get('learning_rate', '0.001')

        self.wandb_monitor = self.kwargs.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs}',
            f'{self.learning_rate}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/BNNBN', exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/BNNBN/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet BNNBN', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        if not torch.cuda.is_available():
            sys.exit(1)
        start_t = time.time()

        cudnn.benchmark = True
        cudnn.enabled = True

        self.model_student.cuda()
        print(f'Student Model:{self.model_student}')

        # load teacher model
        if self.loss_type == 'kd':
            print('Loading teacher model')
            logger.info('Loading teacher model')
            if not 'nfnet' in self.teacher:
                model_teacher = models.__dict__[self.teacher](pretrained=True)
                classes_in_teacher = model_teacher.fc.out_features
                num_features = model_teacher.fc.in_features
            else:
                model_teacher = timm.create_model(self.teacher,
                                                  pretrained=True)
                classes_in_teacher = model_teacher.head.fc.out_features
                num_features = model_teacher.head.fc.in_features

            if not classes_in_teacher == self.num_classes:
                print('* change fc layers in teacher')
                if not 'nfnet' in self.teacher:
                    model_teacher.fc = nn.Linear(num_features,
                                                 self.num_classes)
                else:
                    model_teacher.head.fc = nn.Linear(num_features,
                                                      self.num_classes)
                print('* loading pretrained teacher weight from {}'.format(
                    self.teacher_weight))
                pretrain_teacher = torch.load(self.teacher_weight,
                                              map_location='cpu')['state_dict']
                model_teacher.load_state_dict(pretrain_teacher)

            print('Teacher Model: {}'.format(model_teacher))
            logger.info('Teacher Model: {}'.format(model_teacher))
            model_teacher.cuda()
            for p in model_teacher.parameters():
                p.requires_grad = False
            model_teacher.eval()

        # criterion
        criterion = nn.CrossEntropyLoss().cuda()
        criterion_smooth = CrossEntropyLabelSmooth(self.num_classes,
                                                   self.label_smooth).cuda()
        criterion_kd = DistributionLoss()

        # optimizer
        all_parameters = self.model_student.parameters()
        weight_parameters = []
        for pname, p in self.model_student.named_parameters():
            if p.ndimension() == 4 or 'conv' in pname:
                weight_parameters.append(p)
        weight_parameters_id = list(map(id, weight_parameters))
        other_parameters = list(
            filter(lambda p: id(p) not in weight_parameters_id,
                   all_parameters))

        optimizer = torch.optim.Adam(
            [
                {
                    'params': other_parameters
                },
                {
                    'params': weight_parameters,
                    'weight_decay': self.weight_decay
                },
            ],
            lr=self.learning_rate,
        )

        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lambda step: (1.0 - step / self.epochs), last_epoch=-1)
        start_epoch = 0
        best_top1_acc = 0

        if self.pretrained:
            print('Loading pretrained weight {}'.format(self.pretrained))
            logger.info('Loading pretrained weight {}'.format(self.pretrained))
            pretrain_student = torch.load(self.pretrained)
            if 'state_dict' in pretrain_student.keys():
                pretrain_student = pretrain_student['state_dict']

            for key in pretrain_student.keys():
                if not key in self.model_student.state_dict().keys():
                    print('unload key: {}'.format(key))

            self.model_student.load_state_dict(pretrain_student, strict=False)

        if self.resume:
            checkpoint_tar = os.path.join(self.save, 'checkpoint.pth.tar')
            if os.path.exists(checkpoint_tar):
                print('Loading from the checkpoint {}'.format(checkpoint_tar))
                logger.info(
                    'Loading from the checkpoint {}'.format(checkpoint_tar))
                checkpoint = torch.load(checkpoint_tar)
                start_epoch = checkpoint['epoch']
                best_top1_acc = checkpoint['best_top1_acc']
                self.model_student.load_state_dict(checkpoint['state_dict'])
                optimizer.load_state_dict(checkpoint['optimizer'])
                scheduler.load_state_dict(checkpoint['scheduler'])
                print('Loaded checkpoint {} epoch = {}'.format(
                    checkpoint_tar, checkpoint['epoch']))
                logger.info('Loaded checkpoint {} epoch = {}'.format(
                    checkpoint_tar, ['epoch']))
            else:
                raise ValueError('no checkpoint for resume')

        if self.loss_type == 'kd':
            if not classes_in_teacher == self.num_classes:
                self.validate('teacher', self.val_loader, model_teacher,
                              criterion)

        # train the model
        epoch = start_epoch
        epochs_list = []
        train_top1_acc_list = []
        train_top5_acc_list = []
        val_top1_acc_list = []
        val_top5_acc_list = []

        while epoch < self.epochs:
            if self.loss_type == 'kd':
                train_obj, train_top1_acc, train_top5_acc = self.train_kd(
                    epoch,
                    self.train_loader,
                    self.model_student,
                    model_teacher,
                    criterion_kd,
                    optimizer,
                    scheduler,
                )

                train_top1_acc_list.append(train_top1_acc)
                train_top5_acc_list.append(train_top5_acc.cpu().numpy())
            elif self.loss_type == 'ce':
                train_obj, train_top1_acc, train_top5_acc = self.train(
                    epoch,
                    self.train_loader,
                    self.model_student,
                    criterion,
                    optimizer,
                    scheduler,
                )

                train_top1_acc_list.append(train_top1_acc)
                train_top5_acc_list.append(train_top5_acc)
            elif self.loss_type == 'ls':
                train_obj, train_top1_acc, train_top5_acc = self.train(
                    epoch,
                    self.train_loader,
                    self.model_student,
                    criterion_smooth,
                    optimizer,
                    scheduler,
                )

                train_top1_acc_list.append(train_top1_acc)
                train_top5_acc_list.append(train_top5_acc)
            else:
                raise ValueError('unsupport loss_type')

            valid_obj, valid_top1_acc, valid_top5_acc = self.validate(
                epoch, self.val_loader, self.model_student, criterion)

            epochs_list.append(epoch)
            val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
            val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

            is_best = False
            if valid_top1_acc > best_top1_acc:
                best_top1_acc = valid_top1_acc
                is_best = True

            save_checkpoint(
                {
                    'epoch': epoch,
                    'state_dict': self.model_student.state_dict(),
                    'best_top1_acc': best_top1_acc,
                    'optimizer': optimizer.state_dict(),
                    'scheduler': scheduler.state_dict(),
                },
                is_best,
                self.save,
            )

            epoch += 1

            if self.wandb_monitor:
                wandb.log({'best_top1_acc': best_top1_acc})

        training_time = (time.time() - start_t) / 3600
        print('Total training time = {} hours'.format(training_time))
        print('Best acc = {}'.format(best_top1_acc))
        logger.info('Total training time = {} hours'.format(training_time))
        logger.info('Best acc = {}'.format(best_top1_acc))

        df_data = np.array([
            epochs_list,
            train_top1_acc_list,
            train_top5_acc_list,
            val_top1_acc_list,
            val_top5_acc_list,
        ]).T
        df = pd.DataFrame(
            df_data,
            columns=[
                'Epochs',
                'Train Top1',
                'Train Top5',
                'Validation Top1',
                'Validation Top5',
            ],
        )
        df.to_csv(f'{os.getcwd()}/logs/BNNBN/{self.name}.csv', index=False)

    def train_kd(
        self,
        epoch,
        train_loader,
        model_student,
        model_teacher,
        criterion,
        optimizer,
        scheduler,
    ):
        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        model_student.train()
        model_teacher.eval()
        end = time.time()
        scheduler.step()

        for param_group in optimizer.param_groups:
            cur_lr = param_group['lr']
        print('learning_rate:', cur_lr)

        epoch_iterator = tqdm(
            train_loader,
            desc=
            'Training KD in BNNBN Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, target) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.cuda()
            target = target.cuda()

            # compute outputy
            logits_student = model_student(images)
            logits_teacher = model_teacher(images)
            loss = criterion(logits_student, logits_teacher)

            # measure accuracy and record loss
            prec1, prec5 = accuracy(logits_student, target, topk=(1, 5))
            n = images.size(0)
            losses.update(loss.item(), n)  # accumulated loss
            top1.update(prec1.item(), n)
            top5.update(prec5.item(), n)

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()

            # clip gradient if necessary
            if self.agc:
                parameters_list = []
                for name, p in model_student.named_parameters():
                    if not 'fc' in name:
                        parameters_list.append(p)
                adaptive_clip_grad(parameters_list,
                                   clip_factor=self.clip_value)

            optimizer.step()

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training KD in BNNBN Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            logger.info(
                'Training KD in BNNBN Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss_KD': losses.val,
                    'train_top1_acc_KD': top1.val,
                    'train_top5_acc_KD': top5.val,
                })

        return losses.avg, top1.avg, top5.avg

    def train(self, epoch, train_loader, model_student, criterion, optimizer,
              scheduler):
        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            train_loader,
            desc=
            'Training in BNNBN Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        model_student.train()
        end = time.time()
        scheduler.step()

        for param_group in optimizer.param_groups:
            cur_lr = param_group['lr']

        for i, (images, target) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.cuda()
            target = target.cuda()

            # compute outputy
            logits_student = model_student(images)
            loss = criterion(logits_student, target)

            # measure accuracy and record loss
            prec1, prec5 = accuracy(logits_student, target, topk=(1, 5))
            n = images.size(0)
            losses.update(loss.item(), n)  # accumulated loss
            top1.update(prec1.item(), n)
            top5.update(prec5.item(), n)

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()

            # clip gradient if necessary
            if self.agc:
                parameters_list = []
                for name, p in model_student.named_parameters():
                    if not 'fc' in name:
                        parameters_list.append(p)
                adaptive_clip_grad(parameters_list,
                                   clip_factor=self.clip_value)

            optimizer.step()

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training in BNNBN Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            logger.info(
                'Training in BNNBN Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss': losses.val,
                    'train_top1_acc': top1.val,
                    'train_top5_acc': top5.val,
                })

        return losses.avg, top1.avg, top5.avg

    def validate(self, epoch, val_loader, model, criterion):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            val_loader,
            desc=
            'Validating in BNNBN Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        # switch to evaluation mode
        model.eval()
        with torch.no_grad():
            end = time.time()

            for i, (images, target) in enumerate(epoch_iterator):
                images = images.cuda()
                target = target.cuda()

                # compute output
                logits = model(images)
                loss = criterion(logits, target)

                # measure accuracy and record loss
                pred1, pred5 = accuracy(logits, target, topk=(1, 5))
                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch.nn as nn

stage_repeat = [3, 4, 6, 3]
stage_out_channel = [64] + [256] * 3 + [512] * 4 + [1024] * 6 + [2048] * 3


def adapt_channel(sparsity):
    stage_oup_cprate = []
    stage_oup_cprate += [sparsity[0]]
    for i in range(len(stage_repeat) - 1):
        stage_oup_cprate += [sparsity[i + 1]] * stage_repeat[i]
    stage_oup_cprate += [0.0] * stage_repeat[-1]

    mid_scale_cprate = sparsity[len(stage_repeat):]

    overall_channel = []
    mid_channel = []
    for i in range(len(stage_out_channel)):
        if i == 0:
            overall_channel += [
                int(stage_out_channel[i] * (1 - stage_oup_cprate[i]))
            ]
        else:
            overall_channel += [
                int(stage_out_channel[i] * (1 - stage_oup_cprate[i]))
            ]
            mid_channel += [
                int(stage_out_channel[i] // 4 * (1 - mid_scale_cprate[i - 1]))
            ]

    return overall_channel, mid_channel


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding."""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=3,
                     stride=stride,
                     padding=1,
                     bias=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution."""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=1,
                     stride=stride,
                     bias=False)


class Bottleneck(nn.Module):

    def __init__(self,
                 midplanes,
                 inplanes,
                 planes,
                 stride=1,
                 is_downsample=False):
        super(Bottleneck, self).__init__()
        expansion = 4

        # midplanes = int(planes/expansion)
        norm_layer = nn.BatchNorm2d
        self.conv1 = conv1x1(inplanes, midplanes)
        self.bn1 = norm_layer(midplanes)
        self.relu1 = nn.ReLU(inplace=True)

        self.conv2 = conv3x3(midplanes, midplanes, stride)
        self.bn2 = norm_layer(midplanes)
        self.relu2 = nn.ReLU(inplace=True)

        self.conv3 = conv1x1(midplanes, planes)
        self.bn3 = norm_layer(planes)
        self.relu3 = nn.ReLU(inplace=True)

        self.stride = stride
        self.inplanes = inplanes
        self.planes = planes
        self.midplanes = midplanes

        self.is_downsample = is_downsample
        self.expansion = expansion

        if is_downsample:
            self.downsample = nn.Sequential(
                conv1x1(inplanes, planes, stride=stride),
                norm_layer(planes),
            )

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu2(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.is_downsample:
            identity = self.downsample(x)

        out += identity
        out = self.relu3(out)

        return out


class LambdaLayer(nn.Module):

    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


class ResNet50(nn.Module):

    def __init__(self, sparsity, num_classes=1000):
        super(ResNet50, self).__init__()

        overall_channel, mid_channel = adapt_channel(sparsity)
        self.num_blocks = stage_repeat

        layer_num = 0
        self.conv1 = nn.Conv2d(
            3,
            overall_channel[layer_num],
            kernel_size=7,
            stride=2,
            padding=3,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(overall_channel[layer_num])
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = nn.ModuleList()
        self.layer2 = nn.ModuleList()
        self.layer3 = nn.ModuleList()
        self.layer4 = nn.ModuleList()

        layer_num += 1
        for i in range(len(stage_repeat)):
            if i == 0:
                eval('self.layer%d' % (i + 1)).append(
                    Bottleneck(
                        mid_channel[layer_num - 1],
                        overall_channel[layer_num - 1],
                        overall_channel[layer_num],
                        stride=1,
                        is_downsample=True,
                    ))
                layer_num += 1
            else:
                eval('self.layer%d' % (i + 1)).append(
                    Bottleneck(
                        mid_channel[layer_num - 1],
                        overall_channel[layer_num - 1],
                        overall_channel[layer_num],
                        stride=2,
                        is_downsample=True,
                    ))
                layer_num += 1

            for j in range(1, stage_repeat[i]):
                eval('self.layer%d' % (i + 1)).append(
                    Bottleneck(
                        mid_channel[layer_num - 1],
                        overall_channel[layer_num - 1],
                        overall_channel[layer_num],
                    ))
                layer_num += 1

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(2048, num_classes)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        for i, block in enumerate(self.layer1):
            x = block(x)
        for i, block in enumerate(self.layer2):
            x = block(x)
        for i, block in enumerate(self.layer3):
            x = block(x)
        for i, block in enumerate(self.layer4):
            x = block(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


def resnet_50(sparsity, num_classes):
    return ResNet50(sparsity=sparsity, num_classes=num_classes)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from .resnet import get_resnet_model
from .mobilenetv2_bireal import get_mobilenet as get_mobilenet_bireal
from .mobilenet import get_mobilenet as get_mobilenet_normal
from .resnet_bireal import make_birealnet18, make_birealnet34, make_birealnet50
from .resnet_chip import resnet_50 as resnet_50_chip

# All models_bnnbn
# reactnet-18
from .models_bnnbn.Qa_reactnet_18_bn import birealnet18 as Qa_reactnet_18_bn
from .models_bnnbn.Qa_reactnet_18_none import birealnet18 as Qa_reactnet_18_none
from .models_bnnbn.Qa_reactnet_18_bf import birealnet18 as Qa_reactnet_18_bf

from .models_bnnbn.Qaw_reactnet_18_bn import birealnet18 as Qaw_reactnet_18_bn
from .models_bnnbn.Qaw_reactnet_18_none import birealnet18 as Qaw_reactnet_18_none
from .models_bnnbn.Qaw_reactnet_18_bf import birealnet18 as Qaw_reactnet_18_bf

# reactnet-A
from .models_bnnbn.Qa_reactnet_A_bn import reactnet as Qa_reactnet_A_bn
from .models_bnnbn.Qa_reactnet_A_none import reactnet as Qa_reactnet_A_none
from .models_bnnbn.Qa_reactnet_A_bf import reactnet as Qa_reactnet_A_bf

from .models_bnnbn.Qaw_reactnet_A_bn import reactnet as Qaw_reactnet_A_bn
from .models_bnnbn.Qaw_reactnet_A_none import reactnet as Qaw_reactnet_A_none
from .models_bnnbn.Qaw_reactnet_A_bf import reactnet as Qaw_reactnet_A_bf


class ModelsFactory(object):

    @staticmethod
    def create_model(name,
                     num_classes=100,
                     pretrained=False,
                     version='original',
                     **kwargs):
        """Returns the requested model, ready for training/pruning with the
        specified method.

        Args:
            name: model name 'resnet18','resnet50'
            num_classes: number of classes
        Return:
            model object
        """

        if name in [
                'resnet18',
                'resnet20',
                'resnet32',
                'resnet50',
                'resnet56',
                'resnet101',
                'resnet110',
        ]:
            if version == 'original':
                assert 'insize' in kwargs, 'should provide input size'
                insize = kwargs['insize']
                model = get_resnet_model(name,
                                         num_classes,
                                         insize=insize,
                                         pretrained=pretrained)

            elif version == 'chip':
                assert 'sparsity' in kwargs, 'should provide sparsity for chip'
                model = resnet_50_chip(sparsity=eval(kwargs['sparsity']),
                                       num_classes=num_classes)
            elif version == 'bireal':
                assert 'insize' in kwargs, 'should provide input size'
                insize = kwargs['insize']
                assert 'num_fp' in kwargs, 'should provide num_fp'
                num_fp = kwargs['num_fp']
                if name == 'resnet18':
                    model = make_birealnet18(num_classes=num_classes,
                                             insize=insize,
                                             num_fp=num_fp)
                elif name == 'resnet34':
                    model = make_birealnet34(num_classes=num_classes,
                                             insize=insize,
                                             num_fp=num_fp)
                elif name == 'resnet50':
                    model = make_birealnet50(num_classes=num_classes,
                                             insize=insize,
                                             num_fp=num_fp)
                else:
                    raise Exception(
                        'unknown model {} for BirealNet, available .models_bnnbn are (resnet18, resnet34, resnet50)'
                        .format(name))

        elif name in ['mobilenetv2']:
            if version == 'original':
                model = get_mobilenet_normal(name, num_classes, **kwargs)
            elif version == 'bireal':
                assert 'num_fp' in kwargs, 'should provide num_fp'
                num_fp = kwargs['num_fp']
                model = get_mobilenet_bireal(num_classes, num_fp=num_fp)
            else:
                raise Exception('unknown model {}'.format(name))

        elif name == 'reactnet-18':
            print('* Model = ReActNet-18')
            if kwargs['binary_w']:
                print('* Binarize both activation and weights')
                if kwargs['bn_type'] == 'bn':
                    print('* with BN')
                    model = Qaw_reactnet_18_bn(num_classes=num_classes)
                elif kwargs['bn_type'] == 'none':
                    print('* without BN')
                    model = Qaw_reactnet_18_none(num_classes=num_classes)
                elif kwargs['bn_type'] == 'bf':
                    print('* BN-Free')
                    model = Qaw_reactnet_18_bf(num_classes=num_classes)

            else:
                print('* Binarize only activation')
                if kwargs['bn_type'] == 'bn':
                    print('* with BN')
                    model = Qa_reactnet_18_bn(num_classes=num_classes)
                elif kwargs['bn_type'] == 'none':
                    print('* without BN')
                    model = Qa_reactnet_18_none(num_classes=num_classes)
                elif kwargs['bn_type'] == 'bf':
                    print('* BN-Free')
                    model = Qa_reactnet_18_bf(num_classes=num_classes)

        elif name == 'reactnet-A':
            print('* Model = reactnet-A')
            if kwargs['binary_w']:
                print('* Binarize both activation and weights')
                if kwargs['bn_type'] == 'bn':
                    print('* with BN')
                    model = Qaw_reactnet_A_bn(num_classes=num_classes)
                elif kwargs['bn_type'] == 'none':
                    print('* without BN')
                    model = Qaw_reactnet_A_none(num_classes=num_classes)
                elif kwargs['bn_type'] == 'bf':
                    print('* BN-Free')
                    model = Qaw_reactnet_A_bf(num_classes=num_classes)

            else:
                print('* Binarize only activation')
                if kwargs['bn_type'] == 'bn':
                    print('* with BN')
                    model = Qa_reactnet_A_bn(num_classes=num_classes)
                elif kwargs['bn_type'] == 'none':
                    print('* without BN')
                    model = Qa_reactnet_A_none(num_classes=num_classes)
                elif kwargs['bn_type'] == 'bf':
                    print('* BN-Free')
                    model = Qa_reactnet_A_bf(num_classes=num_classes)

        else:
            raise Exception('unknown model {}'.format(name))

        return model



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import math
from functools import reduce

import torch
import torch.nn as nn
import torch.nn.functional as F

from .base_model import BaseModel
"""MobileNet in PyTorch.

See the paper "MobileNets: Efficient Convolutional Neural Networks for Mobile
Vision Applications" for more details. Code is taken from
https://github.com/kuangliu/pytorch-cifar/blob/master/models/mobilenet.py
"""


class InvertedResidual(nn.Module):
    """Expand + depthwise + pointwise."""

    def __init__(self, in_planes, out_planes, expansion, stride):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        self.is_shortcut = False
        planes = expansion * in_planes
        self.conv1 = nn.Conv2d(in_planes,
                               planes,
                               kernel_size=1,
                               stride=1,
                               padding=0,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        #         self.conv1, self.bn1 = ModuleInjection.make_prunable(self.conv1, self.bn1)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=planes,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes,
                               out_planes,
                               kernel_size=1,
                               stride=1,
                               padding=0,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(out_planes)
        #         self.conv3, self.bn3 = ModuleInjection.make_prunable(self.conv3, self.bn3)

        self.shortcut = nn.Sequential()
        if stride == 1 and in_planes != out_planes:
            self.is_shortcut = True
            conv_module = nn.Conv2d(in_planes,
                                    out_planes,
                                    kernel_size=1,
                                    stride=1,
                                    padding=0,
                                    bias=False)
            bn_module = nn.BatchNorm2d(out_planes)
            #             conv_module, bn_module = ModuleInjection.make_prunable(conv_module, bn_module)
            if hasattr(bn_module, 'is_imp'):
                bn_module.is_imp = True
            self.shortcut = nn.Sequential(conv_module, bn_module)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out = out + self.shortcut(x) if self.stride == 1 else out
        return out

    def get_flops(self, inp):
        flops = 0

        output = self.conv1(inp)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv1.in_channels * self.conv1.out_channels *
                  active_elements_count)
        flops += 4 * output.numel()  # relu + bn
        output = F.relu(output)

        output = self.conv2(output)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv2.in_channels *
                  (self.conv2.out_channels // self.conv2.groups) * 9 *
                  active_elements_count)
        flops += 4 * output.numel()  # relu + bn
        output = F.relu(output)

        output = self.conv3(output)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv3.in_channels * self.conv3.out_channels *
                  active_elements_count)
        flops += 3 * output.numel()  # bn + residual

        if self.is_shortcut:
            r_output = self.shortcut[0](inp)
            active_elements_count = r_output.shape[0] * reduce(
                lambda x, y: x * y, r_output.shape[2:])
            flops += (self.shortcut[0].in_channels *
                      self.shortcut[0].out_channels * active_elements_count)
            flops += 2 * r_output.numel()

        return flops, output


class MobileNetv2(BaseModel):
    # (expansion, out_planes, num_blocks, stride)
    cfg = [
        (1, 16, 1, 1),
        (6, 24, 2, 2),  # NOTE: change stride 2 -> 1 for CIFAR10
        (6, 32, 3, 2),
        (6, 64, 4, 2),
        (6, 96, 3, 1),
        (6, 160, 3, 2),
        (6, 320, 1, 1),
    ]

    def __init__(self, num_classes=10, cfg=None):
        super(MobileNetv2, self).__init__()
        if cfg:
            self.cfg = cfg
        # NOTE: change conv1 stride 2 -> 1 for CIFAR10
        self.conv1 = nn.Conv2d(3,
                               32,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(32)
        #         self.conv1, self.bn1 = ModuleInjection.make_prunable(self.conv1, self.bn1)
        if hasattr(self.bn1, 'is_imp'):
            self.bn1.is_imp = True
        self.layers = self._make_layers(in_planes=32)
        self.conv2 = nn.Conv2d(self.cfg[-1][1],
                               1280,
                               kernel_size=1,
                               stride=2,
                               padding=0,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(1280)
        #         self.conv2, self.bn2 = ModuleInjection.make_prunable(self.conv2, self.bn2)
        if hasattr(self.bn2, 'is_imp'):
            self.bn2.is_imp = True
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)
        self.linear = nn.Linear(1280, num_classes)

    def _make_layers(self, in_planes):
        layers = []
        for expansion, out_planes, num_blocks, stride in self.cfg:
            strides = [stride] + [1] * (num_blocks - 1)
            for stride in strides:
                layers.append(
                    InvertedResidual(in_planes, out_planes, expansion, stride))
                in_planes = out_planes
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layers(out)
        out = F.relu(self.bn2(self.conv2(out)))
        # NOTE: change pooling kernel_size 7 -> 4 for CIFAR10
        # out = F.avg_pool2d(out, 7)
        out = self.avgpool(out)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out

    def get_flops(self, insize):
        flops = 0
        inp = torch.rand(1, 3, insize, insize).to(self.conv1.weight.device)
        output = self.conv1(inp)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv1.in_channels * self.conv1.out_channels * 9 *
                  active_elements_count)
        flops += 4 * output.numel()  # bn + relu
        output = F.relu(output)

        for b in self.layers:
            b_flops, output = b.get_flops(output)
            flops += b_flops

        output = self.conv2(output)
        active_elements_count = output.shape[0] * reduce(
            lambda x, y: x * y, output.shape[2:])
        flops += (self.conv2.in_channels * self.conv2.out_channels *
                  active_elements_count)
        flops += 4 * output.numel()  # bn + relu
        output = F.relu(output)

        ops = output.shape[0] * reduce(lambda x, y: x * y, output.shape[2:])
        output = self.avgpool(output)
        flops += ops * output.numel()

        flops += (1 + self.linear.in_features) * self.linear.out_features
        return flops

    def removable_orphans(self):
        num_removed = 0
        for b in self.layers:
            m1, m2 = b.bn1, b.bn3
            if self.is_all_pruned(m1) or self.is_all_pruned(m2):
                num_removed += self.n_remaining(m1) + self.n_remaining(m2)
        return num_removed

    def remove_orphans(self):
        num_removed = 0
        for b in self.layers:
            m1, m2 = b.bn1, b.bn3
            if self.is_all_pruned(m1) or self.is_all_pruned(m2):
                num_removed += self.n_remaining(m1) + self.n_remaining(m2)
                m1.pruned_zeta.data.copy_(torch.zeros_like(m1.pruned_zeta))
                m2.pruned_zeta.data.copy_(torch.zeros_like(m2.pruned_zeta))
        return num_removed


def get_mobilenet(model, num_classes, cfg=None):
    """Returns the requested model, ready for training/pruning with the
    specified method.

    :param model: str
    :param method: full or prune
    :param num_classes: int, num classes in the dataset
    :return: A prunable MobileNet model
    """
    #     ModuleInjection.pruning_method = method
    #     ModuleInjection.prunable_modules = []
    if model == 'mobilenetv2':
        net = MobileNetv2(num_classes, cfg)
    #     net.prunable_modules = ModuleInjection.prunable_modules
    return net



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy, random, pickle
import torch
import torch.nn as nn
from collections import OrderedDict
from trailmet.utils import seed_everything
from trailmet.algorithms.quantize.quantize import BaseQuantization
from trailmet.models.resnet import BasicBlock, Bottleneck
from trailmet.models.mobilenet import InvertedResidual
from trailmet.algorithms.quantize.qmodel import (
    QBasicBlock,
    QBottleneck,
    QInvertedResidual,
)
from trailmet.algorithms.quantize.methods import BitSplitQuantizer, ActQuantizer

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import AverageMeter, accuracy, save_checkpoint

logger = logging.getLogger(__name__)

global feat, prev_feat, conv_feat


def hook(module, input, output):
    global feat
    feat = output.data.cpu().numpy()


def current_input_hook(module, inputdata, outputdata):
    global prev_feat
    prev_feat = inputdata[0].data


def conv_hook(module, inputdata, outputdata):
    global conv_feat
    conv_feat = outputdata.data


class QuantModel(nn.Module):

    def __init__(self, model: nn.Module, arch='ResNet50'):
        super().__init__()
        self.supported = {
            BasicBlock: QBasicBlock,
            Bottleneck: QBottleneck,
            InvertedResidual: QInvertedResidual,
        }
        if arch == 'ResNet50':
            setattr(model, 'quant', ActQuantizer())
            setattr(model, 'fc', nn.Sequential(ActQuantizer(), model.fc))
        if arch == 'MobileNetV2':
            setattr(model, 'conv2', nn.Sequential(ActQuantizer(), model.conv2))
            setattr(model, 'linear', nn.Sequential(ActQuantizer(),
                                                   model.linear))
        self.quant_block_refactor(model)

    def quant_block_refactor(self, module: nn.Module):
        """Recursively modify the supported conv-blocks to add activation
        quantization layers :param module: nn.Module with supported conv-block
        classes in its children."""
        for name, child_module in module.named_children():
            if type(child_module) in self.supported:
                setattr(module, name,
                        self.supported[type(child_module)](child_module))
            elif isinstance(child_module,
                            (nn.Conv2d, nn.Linear, nn.ReLU, nn.ReLU6)):
                continue
            else:
                self.quant_block_refactor(child_module)


class BitSplit(BaseQuantization):
    """
    Class for post-training quantization using bit-split and stitching method
    based on - Towards accurate post-training network quantization via
    bit-split and stitching [https://dl.acm.org/doi/abs/10.5555/3524938.3525851]

    :param W_BITS: bitwidth for weight quantization
    :param A_BITS: bitwidth for activation quantization
    :param CHANNEL_WISE: apply channel-wise quantization for weights
    :param ACT_QUANT: apply activation quantization
    :param HEAD_STEM_PRECISION: bitwidth for first and last layer
    :param PREC_CONFIG: list of bitwidths of the body for mixed precision
    :param CALIB_BATCHES: num of batches in calibration dataset
    :param LOAD_ACT_SCALES: load precomputed weight scales
    :param LOAD_WEIGHT_SCALES: load precomputed activation scales
    :param SAVE_PATH: path for storing quantized weights and scales
    """

    def __init__(self, model: nn.Module, dataloaders, **kwargs):
        super(BitSplit, self).__init__(**kwargs)
        self.model = model
        self.train_loader = dataloaders['train']
        self.kwargs = kwargs
        self.w_bits = self.kwargs.get('W_BITS', 8)
        self.a_bits = self.kwargs.get('A_BITS', 8)
        self.gpu_id = self.kwargs.get('GPU_ID', 0)
        self.seed = self.kwargs.get('SEED', 42)
        self.device = torch.device('cuda:{}'.format(self.gpu_id))
        torch.cuda.set_device(self.gpu_id)
        seed_everything(self.seed)
        self.save_path = self.kwargs.get('SAVE_PATH', './')
        self.arch = self.kwargs.get('ARCH', '')
        self.dataset = self.kwargs.get('DATASET', '')
        self.precision_config = self.kwargs.get('PREC_CONFIG', [])
        if self.precision_config:
            w_prefix = str(self.precision_config[0]) + '_mix'
        else:
            w_prefix = str(self.w_bits)
        self.prefix = self.save_path + self.arch + '_' + self.dataset + '/W' + w_prefix
        if not os.path.exists(self.prefix):
            os.makedirs(self.prefix)
        self.load_act_scales = self.kwargs.get('LOAD_ACT_SCALES', False)
        self.load_weight_scales = self.kwargs.get('LOAD_WEIGHT_SCALES', False)
        self.calib_batches = self.kwargs.get('CALIB_BATCHES', 8)
        self.act_quant = self.kwargs.get('ACT_QUANT', True)
        self.head_stem_precision = self.kwargs.get('HEAD_STEM_PRECISION', None)

        self.wandb_monitor = self.kwargs.get('WANDB', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([
            self.dataset_name,
            str(self.a_bits),
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/BitSplit', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/BitSplit/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet BitSplit', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        self.model.to(self.device)
        self.qmodel = copy.deepcopy(self.model)
        QuantModel(self.qmodel, self.arch)

        self.act_quant_modules = []
        for m in self.qmodel.modules():
            if isinstance(m, ActQuantizer):
                m.set_bitwidth(self.a_bits)
                self.act_quant_modules.append(m)
        self.act_quant_modules[-1].set_bitwidth(max(8, self.a_bits))
        assert self.arch in ['MobileNetV2', 'ResNet50']

        print('==> Starting weight quantization')
        logger.info('==> Starting weight quantization')
        self.weight_quantizer(load_only=self.load_weight_scales)
        if self.act_quant:
            if self.load_act_scales:
                scales = np.load(self.prefix + '/act_' + str(self.a_bits) +
                                 '_scales.npy')
                for index, q_module in enumerate(self.act_quant_modules):
                    q_module.set_scale(scales[index])
            else:
                print("==> Starting '{}-bit' activation quantization".format(
                    self.a_bits))
                logger.info(
                    "==> Starting '{}-bit' activation quantization".format(
                        self.a_bits))
                self.act_quantizer(self.qmodel,
                                   prefix=self.prefix,
                                   n_batches=self.calib_batches)

        save_checkpoint(
            {
                'state_dict': self.qmodel.module.state_dict(),
            },
            is_best=False,
            save=self.save,
        )
        # save_state_dict(self.qmodel.state_dict(), self.prefix, filename='state_dict.pth')

        print('testing quantized model')
        logger.info('testing quantized model')

        val_top1_acc_list = []
        val_top5_acc_list = []

        valid_loss, valid_top1_acc, valid_top5_acc = self.test(
            self.qmodel, self.dataloaders['val'], nn.CrossEntropyLoss())
        val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
        val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

        df_data = np.array([
            val_top1_acc_list,
            val_top5_acc_list,
        ]).T
        df = pd.DataFrame(
            df_data,
            columns=[
                'Validation Top1',
                'Validation Top5',
            ],
        )
        df.to_csv(
            f'{os.getcwd()}/logs/BitSplit/{self.name}.csv',
            index=False,
        )

    # TODO :  Use functions to process submodules of respective models so that adding new models in future is easier
    def weight_quantizer(self, load_only=False):
        """Find optimum weight quantization scales for ResNet & Mobilenet."""
        #### Quantizer for MobilenetV2 ####
        if self.arch == 'MobileNetV2':
            count = 3
            for i in range(len(self.model.layers)):
                if len(self.model.layers[i].shortcut) > 0:
                    count += 4
                else:
                    count += 3
            pbar = tqdm(total=count)
            layer_to_block = [
                1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 4
            ]
            assert (len(self.precision_config) == 0 or len(
                self.precision_config) == 7), 'config list must be of length 7'
            # quantize first conv layer
            conv = self.model.conv1
            conv_quan = self.qmodel.conv1
            w_bit = self.w_bits
            if self.head_stem_precision is not None:
                w_bit = self.head_stem_precision
            if self.precision_config:
                w_bit = self.precision_config[0]
            if w_bit == 32:
                conv_quan.weight.data.copy_(conv.weight.data)
            else:
                if not load_only:
                    conduct_ofwa(
                        self.train_loader,
                        self.model,
                        self.qmodel,
                        conv,
                        conv_quan,
                        w_bit,
                        self.calib_batches,
                        prefix=self.prefix + '/conv1',
                        device=self.device,
                        ec=False,
                    )
                load_ofwa(conv, conv_quan, prefix=self.prefix + '/conv1')
            pbar.update(1)
            time.sleep(0.1)
            # quantize blocks
            for layer_idx in range(len(self.model.layers)):
                current_layer_pretrained = self.model.layers[layer_idx]
                current_layer_quan = self.qmodel.layers[layer_idx]
                w_bit = (self.precision_config[layer_to_block[layer_idx]]
                         if self.precision_config else self.w_bits)
                skip = w_bit == 32
                pkl_path = self.prefix + '/layer' + str(layer_idx)
                # conv layers
                for idx in range(1, 4):
                    conv = eval('current_layer_pretrained.conv{}'.format(idx))
                    conv_quan = eval('current_layer_quan.conv{}'.format(idx))
                    if skip:
                        conv_quan.weight.data.copy_(conv.weight.data)
                    else:
                        if not load_only:
                            conduct_ofwa(
                                self.train_loader,
                                self.model,
                                self.qmodel,
                                conv,
                                conv_quan,
                                w_bit,
                                self.calib_batches,
                                prefix=pkl_path + '_conv{}'.format(idx),
                                device=self.device,
                                dw=(idx == 2),
                                ec=False,
                            )
                        load_ofwa(conv,
                                  conv_quan,
                                  prefix=pkl_path + '_conv' + str(idx))
                    pbar.update(1)
                    time.sleep(0.1)
                # shortcut layer
                if len(current_layer_pretrained.shortcut) > 0:
                    conv = current_layer_pretrained.shortcut[0]
                    conv_quan = current_layer_quan.shortcut[0]
                    if skip:
                        conv_quan.weight.data.copy_(conv.weight.data)
                    else:
                        if not load_only:
                            conduct_ofwa(
                                self.train_loader,
                                self.model,
                                self.qmodel,
                                conv,
                                conv_quan,
                                w_bit,
                                self.calib_batches,
                                prefix=pkl_path + '_shortcut'.format(idx),
                                device=self.device,
                                ec=False,
                            )
                        load_ofwa(conv,
                                  conv_quan,
                                  prefix=pkl_path + '_shortcut')
                    pbar.update(1)
                    time.sleep(0.1)
            # quantize last conv layer
            conv = self.model.conv2
            conv_quan = self.qmodel.conv2[1]
            if self.precision_config:
                w_bit = self.precision_config[-2]
            if w_bit == 32:
                conv_quan.weight.data.copy_(conv.weight.data)
            else:
                if not load_only:
                    conduct_ofwa(
                        self.train_loader,
                        self.model,
                        self.qmodel,
                        conv,
                        conv_quan,
                        w_bit,
                        self.calib_batches,
                        prefix=self.prefix + '/conv2',
                        device=self.device,
                        ec=False,
                    )
                load_ofwa(conv, conv_quan, prefix=self.prefix + '/conv2')
            pbar.update(1)
            time.sleep(0.1)
            # quantize last linear layer
            conv = self.model.linear
            conv_quan = self.qmodel.linear[1]
            w_bit = self.w_bits
            if self.head_stem_precision is not None:
                w_bit = self.head_stem_precision
            if self.precision_config:
                w_bit = self.precision_config[-1]
            if w_bit == 32:
                conv_quan.weight.data.copy_(conv.weight.data)
            else:
                if not load_only:
                    conduct_ofwa(
                        self.train_loader,
                        self.model,
                        self.qmodel,
                        conv,
                        conv_quan,
                        w_bit,
                        self.calib_batches,
                        prefix=self.prefix + '/linear',
                        device=self.device,
                        ec=False,
                    )
                load_ofwa(conv, conv_quan, prefix=self.prefix + '/linear')
            pbar.update(1)
            pbar.close()

        #### Quantizer for Resnet50 ####
        elif self.arch == 'ResNet50':
            count = 2
            for i in range(1, 5):
                layer = eval('self.model.layer{}'.format(i))
                for j in range(len(layer)):
                    count += 3
                    if layer[j].downsample is not None:
                        count += 1
            pbar = tqdm(total=count)
            # quantize first conv layer
            conv = self.model.conv1
            conv_quan = self.qmodel.conv1
            w_bit = self.w_bits
            if self.head_stem_precision is not None:
                w_bit = self.head_stem_precision
            if self.precision_config:
                w_bit = self.precision_config[0]
            if w_bit == 32:
                conv_quan.weight.data.copy_(conv.weight.data)
            else:
                if not load_only:
                    conduct_ofwa(
                        self.train_loader,
                        self.model,
                        self.qmodel,
                        conv,
                        conv_quan,
                        w_bit,
                        self.calib_batches,
                        prefix=self.prefix + '/conv1',
                        device=self.device,
                        ec=False,
                    )
                load_ofwa(conv, conv_quan, prefix=self.prefix + '/conv1')
            pbar.update(1)
            time.sleep(0.1)
            # quantize blocks
            for layer_idx in range(1, 5):
                current_layer_pretrained = eval(
                    'self.model.layer{}'.format(layer_idx))
                current_layer_quan = eval(
                    'self.qmodel.layer{}'.format(layer_idx))
                w_bit = (self.precision_config[layer_idx]
                         if self.precision_config else self.w_bits)
                skip = w_bit == 32
                for block_idx in range(len(current_layer_pretrained)):
                    current_block_pretrained = current_layer_pretrained[
                        block_idx]
                    current_block_quan = current_layer_quan[block_idx]
                    pkl_path = (self.prefix + '/layer' + str(layer_idx) +
                                '_block' + str(block_idx))
                    # conv layers
                    for idx in range(1, 4):
                        conv = eval(
                            'current_block_pretrained.conv{}'.format(idx))
                        conv_quan = eval(
                            'current_block_quan.conv{}'.format(idx))
                        if skip:
                            conv_quan.weight.data.copy_(conv.weight.data)
                        else:
                            if not load_only:
                                conduct_ofwa(
                                    self.train_loader,
                                    self.model,
                                    self.qmodel,
                                    conv,
                                    conv_quan,
                                    w_bit,
                                    self.calib_batches,
                                    prefix=pkl_path + '_conv{}'.format(idx),
                                    device=self.device,
                                    ec=False,
                                )
                            load_ofwa(conv,
                                      conv_quan,
                                      prefix=pkl_path + '_conv' + str(idx))
                        pbar.update(1)
                        time.sleep(0.1)
                    # downsample
                    if current_block_pretrained.downsample is not None:
                        conv = current_block_pretrained.downsample[0]
                        conv_quan = current_block_quan.downsample[0]
                        if skip:
                            conv_quan.weight.data.copy_(conv.weight.data)
                        else:
                            if not load_only:
                                conduct_ofwa(
                                    self.train_loader,
                                    self.model,
                                    self.qmodel,
                                    conv,
                                    conv_quan,
                                    w_bit,
                                    self.calib_batches,
                                    prefix=pkl_path + '_downsample',
                                    device=self.device,
                                    ec=False,
                                )
                            load_ofwa(conv,
                                      conv_quan,
                                      prefix=pkl_path + '_downsample')
                        pbar.update(1)
                        time.sleep(0.1)
            # quantize last fc layer
            conv = self.model.fc
            conv_quan = self.qmodel.fc[1]
            w_bit = self.w_bits
            if self.head_stem_precision is not None:
                w_bit = self.head_stem_precision
            if self.precision_config:
                w_bit = self.precision_config[-1]
            if w_bit == 32:
                conv_quan.weight.data.copy_(conv.weight.data)
            else:
                if not load_only:
                    conduct_ofwa(
                        self.train_loader,
                        self.model,
                        self.qmodel,
                        conv,
                        conv_quan,
                        w_bit,
                        self.calib_batches,
                        prefix=self.prefix + '/fc',
                        device=self.device,
                        ec=False,
                    )
                load_ofwa(conv, conv_quan, prefix=self.prefix + '/fc')
            pbar.update(1)
            pbar.close()
        else:
            raise NotImplementedError

    # TODO : Write this in a more cleaner way
    def act_quantizer(self, model, prefix, n_batches):
        """Find optimum activation quantization scale for ResNet model based on
        feature map."""

        # train_batches = iter(self.train_loader)
        # per_batch = len(next(train_batches)[1])
        # act_sta_len = (n_batches+1)*per_batch
        def get_safe_len(x):
            x /= 10
            y = 1
            while x >= 10:
                x /= 10
                y *= 10
            return int(y)

        act_sta_len = 3000000
        feat_buf = np.zeros(act_sta_len)
        scales = np.zeros(len(self.act_quant_modules))

        pbar = tqdm(
            self.act_quant_modules,
            desc=
            'Activation quantization, q_module [X] (X / X Steps) (prev_layer_scale=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )
        with torch.no_grad():
            for index, q_module in enumerate(pbar):
                batch_iterator = iter(self.train_loader)
                images, targets = next(batch_iterator)
                images = images.cuda()
                targets = targets.cuda()

                handle = q_module.register_forward_hook(hook)
                model(images)
                feat_len = feat.size
                per_batch = min(get_safe_len(feat_len), 100000)
                n_batches = int(act_sta_len / per_batch)

                repeat = True
                while repeat:
                    repeat = False
                    for batch_idx in range(0, n_batches):
                        pbar.set_description(
                            'Activation quantization, q_module [%d] (%d / %d Steps) (prev_layer_scale=%2.5f)'
                            % (
                                index,
                                batch_idx + 1,
                                n_batches,
                                scales[index - 1],
                            ))
                        images, targets = next(batch_iterator)
                        images = images.cuda(device=self.device,
                                             non_blocking=True)
                        model(images)
                        if q_module.signed:
                            feat_tmp = np.abs(feat).reshape(-1)
                        else:
                            feat_tmp = feat[feat > 0].reshape(-1)
                            if feat_tmp.size < per_batch:
                                per_batch = int(per_batch / 10)
                                n_batches = int(n_batches * 10)
                                repeat = True
                                break
                        np.random.shuffle(feat_tmp)
                        feat_buf[batch_idx * per_batch:(batch_idx + 1) *
                                 per_batch] = feat_tmp[0:per_batch]
                    if not repeat:
                        scales[index] = q_module.init_quantization(feat_buf)
                handle.remove()

        pbar.close()
        np.save(
            os.path.join(prefix, 'act_' + str(self.a_bits) + '_scales.npy'),
            scales)
        for index, q_module in enumerate(self.act_quant_modules):
            q_module.set_scale(scales[index])

    def test(self, model, dataloader, loss_fn):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating network (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        model.eval()
        model.to(self.device)

        with torch.no_grad():
            end = time.time()

            for i, (images, labels) in enumerate(epoch_iterator):
                images = images.to(self.device, dtype=torch.float)
                labels = labels.to(self.device)

                preds = model(images)

                loss = loss_fn(preds, labels)

                pred1, pred5 = accuracy(preds, labels, topk=(1, 5))

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating network (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating network (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))
        return losses.avg, top1.avg, top5.avg


def conduct_ofwa(
    train_loader,
    model_pretrained,
    model_quan,
    conv,
    conv_quan,
    bitwidth,
    n_batches,
    device,
    num_epochs=100,
    prefix=None,
    dw=False,
    ec=False,
):
    # for fc
    if not hasattr(conv, 'kernel_size'):
        W = conv.weight.data  # .cpu()
        W_shape = W.shape
        B_sav, B, alpha = BitSplitQuantizer(W.cpu().numpy(), bitwidth).ofwa()
        # B_sav, B, alpha = ofwa(W.cpu().numpy(), bitwidth)
        with open(prefix + '_fwa.pkl', 'wb') as f:
            pickle.dump({'B': B, 'alpha': alpha}, f, pickle.HIGHEST_PROTOCOL)
        if ec:
            W_r = np.multiply(B, np.expand_dims(alpha, 1)).reshape(W_shape)
            conv_quan.weight.data.copy_(torch.from_numpy(W_r))
        return

    # conv parameters
    kernel_h, kernel_w = conv.kernel_size
    pad_h, pad_w = conv.padding
    stride_h, stride_w = conv.stride

    handle_prev = conv_quan.register_forward_hook(current_input_hook)
    handle_conv = conv.register_forward_hook(conv_hook)

    batch_iterator = iter(train_loader)

    # weights and bias
    W = conv.weight.data  # .cpu()
    if conv.bias is None:
        bias = torch.zeros(W.shape[0]).to(conv.weight.device)
    else:
        bias = conv.bias.data  # .cpu()

    # feat extract
    per_batch = 400
    input, target = next(batch_iterator)
    input_pretrained = input.cuda(device=device, non_blocking=True)
    input_quan = input.cuda(device=device, non_blocking=True)
    model_pretrained(input_pretrained)
    model_quan(input_quan)
    [prev_feat_n, prev_feat_c, prev_feat_h, prev_feat_w] = prev_feat.shape
    [conv_feat_n, conv_feat_c, conv_feat_h, conv_feat_w] = conv_feat.shape

    X = torch.zeros(n_batches * per_batch, prev_feat_c, kernel_h,
                    kernel_w).to(device)
    Y = torch.zeros(n_batches * per_batch, conv_feat_c).to(device)

    for batch_idx in range(0, n_batches):
        input, target = next(batch_iterator)
        input_pretrained = input.cuda(device=device, non_blocking=True)
        model_pretrained(input_pretrained)
        input_quan = input.cuda(device=device, non_blocking=True)
        model_quan(input_quan)

        prev_feat_pad = torch.zeros(prev_feat_n, prev_feat_c,
                                    prev_feat_h + 2 * pad_h,
                                    prev_feat_w + 2 * pad_w).to(device)
        prev_feat_pad[:, :, pad_h:pad_h + prev_feat_h,
                      pad_w:pad_w + prev_feat_w] = prev_feat
        prev_feat_pad = (prev_feat_pad.unfold(2, kernel_h, stride_h).unfold(
            3, kernel_w, stride_w).permute(0, 2, 3, 1, 4, 5))
        [
            feat_pad_n,
            feat_pad_h,
            feat_pad_w,
            feat_pad_c,
            feat_pad_hh,
            feat_pad_ww,
        ] = prev_feat_pad.shape
        assert feat_pad_hh == kernel_h
        assert feat_pad_ww == kernel_w

        prev_feat_pad = prev_feat_pad.reshape(
            feat_pad_n * feat_pad_h * feat_pad_w, feat_pad_c, kernel_h,
            kernel_w)
        rand_index = list(range(prev_feat_pad.shape[0]))
        random.shuffle(rand_index)
        rand_index = rand_index[0:per_batch]
        X[per_batch * batch_idx:per_batch *
          (batch_idx + 1), :] = prev_feat_pad[rand_index, :]
        conv_feat_tmp = conv_feat.permute(0, 2, 3, 1).reshape(
            -1, conv_feat_c) - bias
        Y[per_batch * batch_idx:per_batch *
          (batch_idx + 1), :] = conv_feat_tmp[rand_index, :]

    handle_prev.remove()
    handle_conv.remove()

    ## ofwa init
    W_shape = W.shape
    X = X.cpu().numpy()
    Y = Y.cpu().numpy()
    W = W.reshape(W_shape[0], -1)
    if dw:
        B, alpha = BitSplitQuantizer(W.cpu().numpy(),
                                     bitwidth).ofwa_rr_dw(X, Y, num_epochs)
    else:
        B, alpha = BitSplitQuantizer(W.cpu().numpy(),
                                     bitwidth).ofwa_rr(X, Y, num_epochs)
    with open(prefix + '_rr_b30x400_e100.pkl', 'wb') as f:
        pickle.dump({'B': B, 'alpha': alpha}, f, pickle.HIGHEST_PROTOCOL)


def load_ofwa(conv, conv_quan, prefix=None):
    # for fc
    if not hasattr(conv, 'kernel_size'):
        W = conv.weight.data  # .cpu()
        W_shape = W.shape
        with open(prefix + '_fwa.pkl', 'rb') as f:
            B_alpha = pickle.load(f)
            B = B_alpha['B']
            alpha = B_alpha['alpha']
        W_r = np.multiply(B, np.expand_dims(alpha, 1)).reshape(W_shape)
        conv_quan.weight.data.copy_(torch.from_numpy(W_r))
        return

    # weights and bias
    W = conv.weight.data  # .cpu()
    W_shape = W.shape

    with open(prefix + '_rr_b30x400_e100.pkl', 'rb') as f:
        B_alpha = pickle.load(f)
        B = B_alpha['B']
        alpha = B_alpha['alpha']
    W_r = np.multiply(B, np.expand_dims(alpha, 1)).reshape(W_shape)
    conv_quan.weight.data.copy_(torch.from_numpy(W_r))


def save_state_dict(state_dict, path, filename='state_dict.pth'):
    saved_path = os.path.join(path, filename)
    new_state_dict = OrderedDict()
    for key in state_dict.keys():
        if '.module.' in key:
            new_state_dict[key.replace('.module.',
                                       '.')] = state_dict[key].cpu()
        else:
            new_state_dict[key] = state_dict[key].cpu()
    torch.save(new_state_dict, saved_path)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
import json
from enum import Enum
from functools import reduce
from itertools import chain
from typing import Dict, List

import pandas as pd
from torch.nn import Conv2d, BatchNorm2d, Linear

from .utils import (
    cal_threshold_by_bn2d_weights,
    mask2idxes,
    top_k_idxes,
    prune_conv2d,
    prune_fc,
    prune_bn2d,
    ceil,
    round_up_to_power_of_2,
)

SHORTCUTS_MERGE_OR = 'or'
SHORTCUTS_MERGE_AND = 'and'

__all__ = [
    'ChannelRounding',
    'Conv2dWrapper',
    'LinearWrapper',
    'BN2dWrapper',
    'SlimPruner',
]


class ChannelRounding(Enum):
    NONE = 'none'
    EIGHT = 'eight'  # 8 的倍数
    TWO_POW = 'two_pow'  # 2^n


class Conv2dWrapper:

    def __init__(self, module, name, prev_bn_name=None, next_bn_name=None):
        self.module: Conv2d = copy.deepcopy(module)
        self.pruned_module: Conv2d = module
        self.name = name
        self.prev_bn_name = prev_bn_name
        self.next_bn_name = next_bn_name

        self.is_pruned = False
        self.in_channels_keep_idxes = None
        self.out_channels_keep_idxes = None

    def prune(self, prev_bn=None, next_bn=None):
        if prev_bn is not None:
            assert prev_bn.is_pruned
        if next_bn is not None:
            assert next_bn.is_pruned

        self.prune_by_idxes(
            prev_bn.keep_idxes if prev_bn else None,
            next_bn.keep_idxes if next_bn else None,
        )

    def prune_by_idxes(self, prev_bn_keep_idxes=None, next_bn_keep_idxes=None):
        self.in_channels_keep_idxes, self.out_channels_keep_idxes = prune_conv2d(
            self.pruned_module, prev_bn_keep_idxes, next_bn_keep_idxes)

        self.is_pruned = True

    def prune_info(self):
        in_channels = self.module.in_channels
        out_channels = self.module.out_channels
        pruned_in_channels = self.pruned_module.in_channels
        pruned_out_channels = self.pruned_module.out_channels

        return {
            'name':
            self.name,
            'weight shape':
            f'[{in_channels},{out_channels}] g={self.module.groups}',
            'pruned weight shape':
            f'[{pruned_in_channels},{pruned_out_channels}] g={self.pruned_module.groups}',
            'prune percent':
            f'{100 - (pruned_in_channels * pruned_out_channels) / (in_channels * out_channels) * 100:.2f}%',
        }

    def prune_result(self):
        return {
            'name': self.name,
            'in_channels_keep_idxes': self.in_channels_keep_idxes,
            'out_channels_keep_idxes': self.out_channels_keep_idxes,
        }


class LinearWrapper:

    def __init__(self, module, name, prev_bn_name=None):
        self.module = copy.deepcopy(module)
        self.pruned_module: Linear = module
        self.name = name
        self.prev_bn_name = prev_bn_name
        self.is_pruned = False
        self.in_features_keep_idxes = None

    def prune(self, prev_bn):
        """
        Args:
            prev_bn:

        Returns:

        """
        self.in_features_keep_idxes = prune_fc(self.pruned_module,
                                               prev_bn.keep_idxes,
                                               prev_bn.module.num_features)
        self.is_pruned = True

    def prune_info(self):
        in_features = self.module.in_features
        pruned_in_features = self.pruned_module.in_features
        return {
            'name': self.name,
            'in_features': f'{pruned_in_features}/{in_features}',
            'prune percent': f'{100 - pruned_in_features/in_features*100:.2f}',
        }

    def prune_result(self):
        return {
            'name': self.name,
            'in_features_keep_idxes': self.in_features_keep_idxes,
        }


class BN2dWrapper:

    def __init__(self, module: BatchNorm2d, name):
        self.module = copy.deepcopy(module)
        self.pruned_module = module
        self.name = name
        self.keep_idxes = None
        self.is_pruned = False

    @property
    def is_idxes_calculated(self):
        return self.keep_idxes is not None

    def in_channels(self) -> int:
        return self.module.num_features

    def cal_keep_idxes(
        self,
        threshold: float,
        min_keep_ratio: float = 0,
        channel_rounding: ChannelRounding = ChannelRounding.NONE,
    ):
        """根据所有 BatchNorm2d 层的 gamma 系计算每层 bn 的 keep_idxes."""
        mask = self.module.weight.data.abs().gt(threshold).cpu().numpy()
        idxes = mask2idxes(mask)

        if len(idxes) == 0:
            if min_keep_ratio == 0:
                raise RuntimeError('')
            else:
                idxes = top_k_idxes(self.module, min_keep_ratio)

        if channel_rounding == ChannelRounding.EIGHT:
            k = ceil(len(idxes), 8)
            if k != len(idxes):
                print(f'{self.name} channel round up: {len(idxes)} -> {k}')
            idxes = top_k_idxes(self.module, k=k)
        elif channel_rounding == ChannelRounding.TWO_POW:
            k = round_up_to_power_of_2(len(idxes))
            if k != len(idxes):
                print(f'{self.name} channel round up: {len(idxes)} -> {k}')
            idxes = top_k_idxes(self.module, k=k)

        self.keep_idxes = idxes

    def set_fixed_ratio(self, keep_ratio: float = 1):
        idxes = top_k_idxes(self.module, keep_ratio)
        self.keep_idxes = idxes

    def prune(self):
        assert self.keep_idxes is not None
        prune_bn2d(self.pruned_module, self.keep_idxes)
        self.is_pruned = True

    def prune_info(self):
        str1 = f'{len(self.keep_idxes)}/{self.in_channels()}'
        str2 = (
            f'{(self.in_channels()-len(self.keep_idxes))/self.in_channels()*100:.2f}%'
        )
        return {'name': self.name, 'channels': str1, 'prune percent': str2}

    def prune_result(self):
        return {'name': self.name, 'keep_idxes': self.keep_idxes}


class SlimPruner:
    PRUNING_RESULT_KEY = '_slim_pruning_result'

    def __init__(self, model, schema: str = None):
        """

        Args:
            model: Module will be modified inplace
            schema: pruning schema
        """
        self.pruned_model = model

        if schema is not None:
            if isinstance(schema, str):
                with open(schema, 'r') as f:
                    schema = json.loads(f.read())

            self._add_prefix_to_config_name(schema)
            self.channel_rounding = schema.get('channel_rounding', 'none')

            modules = {}
            for it in schema['modules']:
                modules[it['name']] = it

            self.conv2d_modules = {}
            self.bn2d_modules = {}
            self.fc_modules = {}
            self.cats = schema.get('cats', [])
            self.shortcuts = schema.get('shortcuts', [])
            self.depthwise_conv_adjacent_bn = schema.get(
                'depthwise_conv_adjacent_bn', [])
            self.fixed_bn_ratio = schema.get('fixed_bn_ratio', [])
            print
            for name, module in self.pruned_model.named_modules():
                if name != list(modules.keys())[0].split('.')[0]:
                    name = 'model.' + name
                if isinstance(module, Conv2d):
                    self.conv2d_modules[name] = Conv2dWrapper(
                        module,
                        modules[name]['name'],
                        prev_bn_name=modules[name].get('prev_bn', ''),
                        next_bn_name=modules[name].get('next_bn', ''),
                    )
                elif isinstance(module, Linear):
                    if name in modules:
                        self.fc_modules[name] = LinearWrapper(
                            module,
                            name,
                            prev_bn_name=modules[name].get('prev_bn', ''))
                elif isinstance(module, BatchNorm2d):
                    self.bn2d_modules[name] = BN2dWrapper(module, name)

    def _add_prefix_to_config_name(self, config):
        prefix = config.get('prefix', '')
        for it in config['modules']:
            it['name'] = prefix + it['name']
            if 'prev_bn' in it and it['prev_bn']:
                it['prev_bn'] = prefix + it['prev_bn']
            if 'next_bn' in it and it['next_bn']:
                it['next_bn'] = prefix + it['next_bn']
        """
        "shortcuts": [
            {
                "names": [
                    "bn1",
                    "layer1.0.bn2",
                    "layer1.1.bn2"
                ],
                "method": "or"
            }
        ]
        """
        for it in config.get('shortcuts', []):
            it['names'] = [prefix + _ for _ in it['names']]
        """
        "cats": [
            {
                "names": [ "bn1", "bn2"],
            }
        ]
        """
        for it in config.get('cats', []):
            it['input_bn_names'] = [prefix + _ for _ in it['input_bn_names']]
            it['output_conv_names'] = [
                prefix + _ for _ in it['output_conv_names']
            ]

        for it in config.get('depthwise_conv_adjacent_bn', []):
            it['names'] = [prefix + _ for _ in it['names']]

    def run(self, ratio: float):
        threshold = cal_threshold_by_bn2d_weights(
            [it.module for it in self.bn2d_modules.values()], ratio)

        bn2d_prune_info = []
        for name, bn2d in self.bn2d_modules.items():
            bn2d.cal_keep_idxes(
                threshold,
                min_keep_ratio=0.02,
                channel_rounding=ChannelRounding(self.channel_rounding),
            )

        self._apply_fix_bn_ratio()

        bns_should_be_aligned = copy.deepcopy(self.shortcuts)
        remin_depthwise_conv_adjacent_bn = []
        for item1 in self.depthwise_conv_adjacent_bn:
            """{ "names": ["bn1", "bn2"] "method": "or" / "and" }"""
            _merged = False
            for item2 in bns_should_be_aligned:
                if len(set(item1['names']) & set(item2['names'])) != 0:
                    item2['names'].extend(item1['names'])
                    item2['names'] = list(set(item2['names']))
                    _merged = True
                    # TODO: handle "method"
                    break

            if not _merged:
                remin_depthwise_conv_adjacent_bn.append(item1)

        bns_should_be_aligned.extend(remin_depthwise_conv_adjacent_bn)
        self._align_bns(bns_should_be_aligned,
                        min_keep_ratio=0.05,
                        log_name='shortcuts')

        for bn2d in self.bn2d_modules.values():
            bn2d.prune()
            bn2d_prune_info.append(bn2d.prune_info())

        df = pd.DataFrame(bn2d_prune_info)
        print('\nBatchNorm2d prune info')
        print(df.to_markdown() + '\n')

        convs_after_cat = self._collect_conv_after_cat()
        conv2d_prune_info = []
        # fmt: off
        for conv2d in self.conv2d_modules.values():
            if conv2d.name in convs_after_cat:
                conv2d.prune(
                    None,
                    self.bn2d_modules[conv2d.next_bn_name] if conv2d.next_bn_name else None,
                )
            else:
                conv2d.prune(
                    self.bn2d_modules[conv2d.prev_bn_name] if conv2d.prev_bn_name else None,
                    self.bn2d_modules[conv2d.next_bn_name] if conv2d.next_bn_name else None,
                )
            conv2d_prune_info.append(conv2d.prune_info())
        # fmt: on

        cat_conv_prune_info = self._prune_cat_conv_prev_bn()

        df = pd.DataFrame(conv2d_prune_info)
        print('\nConv2d prune info')
        print(df.to_markdown() + '\n')

        df = pd.DataFrame(cat_conv_prune_info)
        print('\nConv2d after cat prune info')
        print(df.to_markdown() + '\n')

        fc_prune_info = []
        for linear in self.fc_modules.values():
            if not linear.prev_bn_name:
                continue
            linear.prune(self.bn2d_modules[linear.prev_bn_name])
            fc_prune_info.append(linear.prune_info())
        if len(fc_prune_info):
            df = pd.DataFrame(fc_prune_info)
            print('\nLinear prune info')
            print(df.to_markdown() + '\n')

        return self._export_pruning_result()

    def apply_pruning_result(self, pruning_result: List[Dict]):
        info = {it['name']: it for it in pruning_result}

        for name, module in self.pruned_model.named_modules():
            if name not in info:
                continue
            if isinstance(module, Conv2d):
                prune_conv2d(
                    module,
                    info[name]['in_channels_keep_idxes'],
                    info[name]['out_channels_keep_idxes'],
                )

            elif isinstance(module, Linear):
                prune_fc(module, info[name]['in_features_keep_idxes'])
            elif isinstance(module, BatchNorm2d):
                prune_bn2d(module, info[name]['keep_idxes'])

    def _export_pruning_result(self):
        prune_result = []
        for it in chain(
                self.bn2d_modules.values(),
                self.conv2d_modules.values(),
                self.fc_modules.values(),
        ):
            if it.is_pruned:
                prune_result.append(it.prune_result())

        # return {self.PRUNING_RESULT_KEY: prune_result}
        return prune_result

    def _align_bns(self, bn_groups, min_keep_ratio: float, log_name: str):
        """Bn layer is changed inplace [ { "names": ["bn1", "bn2"] "method":
        "or" / "and" } ]

        Returns:
        """
        merged = []
        for i, shortcuts in enumerate(bn_groups):
            assert 'method' in shortcuts
            assert shortcuts['method'] in [
                SHORTCUTS_MERGE_OR, SHORTCUTS_MERGE_AND
            ]

            merge_func = {
                SHORTCUTS_MERGE_AND: lambda x, y: set(x) & set(y),
                SHORTCUTS_MERGE_OR: lambda x, y: set(x) | set(y),
            }[shortcuts['method']]

            bn2d_layers = []
            bn2d_names = shortcuts['names']
            print(f'============{log_name} [{i}]===========')
            for bn2d_name in bn2d_names:
                assert (bn2d_name in self.bn2d_modules
                        ), f'{bn2d_name} is not a BatchNorm2d layer'
                # 防止 bn 出现在多个 shortcuts 中
                assert bn2d_name not in merged

                bn2d = self.bn2d_modules[bn2d_name]
                assert bn2d.is_idxes_calculated
                bn2d_layers.append(bn2d)
                merged.append(bn2d_name)
                print(f'{bn2d_name}: {len(bn2d.keep_idxes)}')

            merged_idxes = list(
                reduce(
                    merge_func,
                    [it.keep_idxes for it in bn2d_layers],
                ))
            print(f'merged indexes length: {len(merged_idxes)}')

            for bn2d in bn2d_layers:
                _merged_idxes = merged_idxes
                if len(merged_idxes) == 0:
                    _merged_idxes = top_k_idxes(bn2d.module, min_keep_ratio)
                bn2d.keep_idxes = _merged_idxes

    def _apply_fix_bn_ratio(self):
        for it in self.fixed_bn_ratio:
            name = it['name']
            ratio = it['ratio']
            assert 0 < ratio < 1
            if isinstance(name, str):
                name = [name]
            for _name in name:
                if _name in self.bn2d_modules:
                    self.bn2d_modules[_name].set_fixed_ratio(1 - ratio)

    def _collect_conv_after_cat(self) -> List[str]:
        if not self.cats:
            return []

        res = []
        for cat_group in self.cats:
            output_conv_names = cat_group['output_conv_names']
            for conv_name in output_conv_names:
                if conv_name not in self.conv2d_modules:
                    raise RuntimeError(
                        f'{conv_name} not exist in {self.conv2d_modules.keys()}'
                    )
            res.extend(output_conv_names)
        return res

    def _prune_cat_conv_prev_bn(self) -> List[Dict]:
        if not self.cats:
            return []

        conv2d_prune_info = []

        # 1. 收集 bn2d_modules 的 keep indexes，cat 起来
        # 2. 剪后续卷积的 kernel
        for cat_group in self.cats:
            input_bn_names = cat_group['input_bn_names']
            output_conv_names = cat_group['output_conv_names']

            final_bn_keep_idxes = []
            idx_offset = 0
            for bn_name in input_bn_names:
                bn_module = self.bn2d_modules.get(bn_name, None)
                if bn_module is None:
                    raise RuntimeError(
                        f'{bn_name} not exist in {self.bn2d_modules.keys()}')

                keep_idxes_offset = [
                    it + idx_offset for it in bn_module.keep_idxes
                ]
                final_bn_keep_idxes.extend(keep_idxes_offset)
                idx_offset += bn_module.in_channels()

            for conv_name in output_conv_names:
                if conv_name not in self.conv2d_modules:
                    raise RuntimeError(
                        f'{conv_name} not exist in {self.conv2d_modules.keys()}'
                    )

                self.conv2d_modules[conv_name].prune_by_idxes(
                    prev_bn_keep_idxes=final_bn_keep_idxes)
                conv2d_prune_info.append(
                    self.conv2d_modules[conv_name].prune_info())

        return conv2d_prune_info



#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# importing the required packages
import sys
import torchvision
from .dataset import BaseDataset


class CIFAR10Dataset(BaseDataset):
    """
    `CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    The CIFAR-10 dataset consists of 60000 32x32 colour images
    in 10 classes, with 6000 images per class. There are 50000
    training images and 10000 test images.The classes are completely
    mutually exclusive.

    References
    ----------
    Learning Multiple Layers of Features from Tiny Images,
    Alex Krizhevsky, 2008.

    Parameters
    ----------
        name (string): dataset name 'CIFAR10', 'CIFAR100', default=None.
        root (string): Root directory where ``cifar-10-batches-py`` exists or will be saved if download flag is set to
        True (default is None).
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set, default=None.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``. Default=None.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it, default=None.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again, default=True.
        split_types (list): the possible values of this parameter are "train", "test" and "val".
            If the split_type contains "val", then shuffle has to be True, default value is None.
        val_fraction (float): If float, should be between 0.0 and 1.0 and represent
        the proportion of the dataset to include in the val split.
        shuffle (bool): Whether or not to shuffle the data before splitting into val from train,
            default is True. If shuffle is true, there should be 'val' in split_types.
        random_seed (int): RandomState instance, default=None.
    """

    def __init__(
        self,
        name=None,
        root=None,
        transform=None,
        target_transform=None,
        download=True,
        split_types=None,
        val_fraction=0.2,
        shuffle=True,
        random_seed=None,
    ):
        super(CIFAR10Dataset, self).__init__(
            name=name,
            root=root,
            transform=transform,
            target_transform=target_transform,
            download=download,
            split_types=split_types,
            val_fraction=val_fraction,
            shuffle=shuffle,
            random_seed=random_seed,
        )

        dataset = torchvision.datasets.CIFAR10
        self.dataset_dict = {}

        for item in self.split_types:
            dataset_type = item
            if item == 'val' and not self.val_exists:
                self.dataset_dict[dataset_type] = None
            data = dataset(
                root=self.root,
                train=(item != 'test'),
                transform=self.transform[item],
                target_transform=self.target_transform[item],
                download=self.download,
            )
            self.dataset_dict[dataset_type] = data

    def build_dict_info(self):
        """
        Behavior:
            This function creates info key in the output dictionary. The info key contains details related to the size
            of the training, validation and test datasets. Further, it can be used to define any additional information
            necessary for the user.
        Returns:
            dataset_dict (dict): Updated with info key that contains details related to the data splits
        """
        self.dataset_dict['info'] = {}
        self.dataset_dict['info']['train_size'] = len(
            self.dataset_dict['train_sampler'])
        self.dataset_dict['info']['val_size'] = len(
            self.dataset_dict['val_sampler'])
        self.dataset_dict['info']['test_size'] = len(self.dataset_dict['test'])
        self.dataset_dict['info']['note'] = (
            'Note that we use the CIFAR10 instance of torchvision for train and validation, '
            'due to which the length of these will be displayed as 50000 when len() is invoked.'
            'For accurate details, extract information from the info keyword in this dict '
        )
        return self.dataset_dict


class CIFAR100Dataset(BaseDataset):
    """
    `CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    The CIFAR-100 dataset consists of 60000 32x32 colour images
    in 100 classes, with 600 images per class. There are 50000
    training images and 10000 test images. The classes are completely
    mutually exclusive.

    References
    ----------
    Learning Multiple Layers of Features from Tiny Images,
    Alex Krizhevsky, 2008.

    Parameters
    ----------
        name (string): dataset name 'CIFAR10', 'CIFAR100', default=None.
        root (string): Root directory of dataset where directory
            ``cifar-10-batches-py`` exists or will be saved to if download
             is set to True, default=None.
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set, default=None.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``. Default=None.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it, default=None.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again, default=True.
        split_types (list): the possible values of this parameter are "train", "test" and "val".
            If the split_type contains "val", then suffle has to be True, default value is None.
        val_fraction (float): If float, should be between 0.0 and 1.0 and represent
        the proportion of the dataset to include in the val split.
        shuffle (bool): Whether or not to shuffle the data before splitting into val from train,
            default is True. If shuffle is true, there should be 'val' in split_types.
        random_seed (int): RandomState instance, default=None.
    """

    def __init__(
        self,
        name=None,
        root=None,
        transform=None,
        target_transform=None,
        download=True,
        split_types=None,
        val_fraction=0.2,
        shuffle=True,
        random_seed=None,
    ):
        super(CIFAR100Dataset, self).__init__(
            name=name,
            root=root,
            transform=transform,
            target_transform=target_transform,
            download=download,
            split_types=split_types,
            val_fraction=val_fraction,
            shuffle=shuffle,
            random_seed=random_seed,
        )

        ## To do: check val_frac is float, else raise error
        ## To do: if shuffle is true, there should be 'val' in train test split
        dataset = torchvision.datasets.CIFAR100
        self.dataset_dict = {}

        for item in self.split_types:
            dataset_type = item
            if item == 'val' and not self.val_exists:
                self.dataset_dict[dataset_type] = None
            data = dataset(
                root=self.root,
                train=(item != 'test'),
                transform=self.transform[item],
                target_transform=self.target_transform[item],
                download=self.download,
            )
            self.dataset_dict[dataset_type] = data

    def build_dict_info(self):
        """
        Behavior:
            This function creates info key in the output dictionary. The info key contains details related to the size
            of the training, validation and test datasets. Further, it can be used to define any additional information
            necessary for the user.
        Returns:
            dataset_dict (dict): Updated with info key that contains details related to the data splits
        """
        self.dataset_dict['info'] = {}
        self.dataset_dict['info']['train_size'] = len(
            self.dataset_dict['train_sampler'])
        self.dataset_dict['info']['val_size'] = len(
            self.dataset_dict['val_sampler'])
        self.dataset_dict['info']['test_size'] = len(self.dataset_dict['test'])
        self.dataset_dict['info']['note'] = (
            'Note that we use the CIFAR100 instance of torchvision for train and validation, '
            'due to which the length of these will be displayed as 50000 when len() is invoked.'
            'For accurate details, extract information from the info keyword in this dict '
        )
        return self.dataset_dict



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torchvision
from .dataset import BaseDataset


class ImageNetDataset(BaseDataset):
    """
    ImageNet <http://image-net.org/>`_ 2012 Classification Dataset.

    Parameters
    ----------
        name (string): dataset name 'ImageNet', default=None.
        root (string): Root directory of dataset or will be saved to if download
             is set to True, default=None.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``. Default=None.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it, default=None.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again, default=True.
        split_types (list): the possible values of this parameter are "train", "test" and "val".
            If the split_type contains "val", then suffle has to be True, default value is None.
        val_fraction (float): If float, should be between 0.0 and 1.0 and represent
        the proportion of the dataset to include in the val split.
        shuffle (bool): Whether or not to shuffle the data before splitting into val from train,
            default is True. If shuffle is true, there should be 'val' in split_types.
        random_seed (int): RandomState instance, default=None.
    """

    def __init__(
        self,
        name=None,
        root=None,
        transform=None,
        target_transform=None,
        download=True,
        split_types=None,
        val_fraction=0.2,
        shuffle=True,
        random_seed=None,
    ):
        super(ImageNetDataset, self).__init__(
            name=name,
            root=root,
            transform=transform,
            target_transform=target_transform,
            download=download,
            split_types=split_types,
            val_fraction=val_fraction,
            shuffle=shuffle,
            random_seed=random_seed,
        )

        ## To do: check val_frac is float, else raise error
        ## To do: if shuffle is true, there should be 'val' in train test split
        dataset = torchvision.datasets.ImageNet
        self.dataset_dict = {}
        for item in self.split_types:
            dataset_type = item + '_dataset'
            if item == 'test':
                data = dataset(
                    root=self.root,
                    split='val',
                    transform=self.transform[item],
                    target_transform=self.target_transform[item],
                )
            else:
                data = dataset(
                    root=self.root,
                    split='train',
                    transform=self.transform[item],
                    target_transform=self.target_transform[item],
                )
            self.dataset_dict[dataset_type] = data
